package week03

import org.scalatest.concurrent.{Signaler, TimeLimitedTests}
import org.scalatest.flatspec.AnyFlatSpec
import org.scalatest.time.{Millis, Span}

import scala.util.Random

class GyakorlatTeszt extends AnyFlatSpec with TimeLimitedTests {
  val timeLimit: Span = Span(1000, Millis)
  override val defaultTestSignaler: Signaler = new Signaler {
    override def apply(testThread: Thread): Unit = {
      println("Ez a teszt túl sokáig fut.")
      testThread.stop()
    } //unsafe, never használd.
  }


  "increase" should "növelnie kéne a lista minden elemét c-vel" in { for ( i <- 1 to 1000 ) {
    val c = Random.nextInt(10) - 5
    def generateTestCase: List[(Int,Int)] = Random.nextInt(10) match {
      case 0 => Nil
      case _ => { val head = Random.nextInt(100); (head,head+c)::generateTestCase }
    }
    val combinedList = generateTestCase
    assert( Gyakorlat.increase( combinedList.map{ _._1 },c) == combinedList.map{ _._2 }, "teszteset: " + combinedList.map{_._1} + ", c=" + c )
  }}

  "positives" should "pontosan a pozitív elemeket kellene megtartsa" in {
    def generateTestCase: (List[Int],List[Int]) = Random.nextInt(10) match {
      case 0 => (Nil,Nil)
      case _ => {
        val head = Random.nextInt( 20 ) - 10
        val (list1, list2) = generateTestCase
        if( head > 0 ) (head::list1, head::list2) else (head::list1, list2)
      }
    }
    for (i <- 1 to 1000) {
      val (test, result) = generateTestCase
      assert( Gyakorlat.positives( test ) == result , "teszteset: " + test )
  }}

  "kovet loptak" should "azt kellene visszaadnia, hogy nagyobb-egyenlő-e a lista összege a célszámnál" in {
    def generateTestCase: (List[Int], Int) = Random.nextInt(10) match {
      case 0 => (Nil,0)
      case _ => {
        val head = Random.nextInt(10)
        val (tail, sum) = generateTestCase
        (head::tail, sum+head)
      }
    }
    for( i <- 1 to 1000 ) {
      val diff = Random.nextInt(20) - 10
      val (test, sum) = generateTestCase
      assert(Gyakorlat.kovetLoptak(test, sum + diff) == (diff <= 0), "teszteset: " + test + ", c=" + (sum+diff) )
    }
  }

  "sumPositives" should "azt kellene visszaadja, hogy mennyi a lista pozitív elemeinek az üsszege" in {
    def generateTestCase: (List[Int], Int) = Random.nextInt(10) match {
      case 0 => (Nil,0)
      case _ => {
        val head = Random.nextInt(20) - 10
        val (tail, sum) = generateTestCase
        if( head > 0 ) (head::tail, sum+head) else (head::tail, sum)
      }
    }
    for( i <- 1 to 1000 ) {
      val (testCase, result) = generateTestCase
      assert(Gyakorlat.sumPositives(testCase) == result, "teszteset: " + testCase )
    }
  }

  "viraglocsolas" should "hányat fordul a viráglocsolóember?" in {
    def generateTestCase(capacity: Int): (List[Int], Int ) = Random.nextInt(10) match {
      case 0 => (Nil, 0)
      case _ => {
        val head = Random.nextInt( capacity * 5 )
        val (tail, sum) = generateTestCase(capacity)
        (head::tail, sum + (head + capacity - 1)/capacity)
      }
    }
    for( i <- 1 to 1000 ) {
      val capacity = Random.nextInt( 10 ) + 1
      val (testCase, result) = generateTestCase(capacity)
      assert(Gyakorlat.viragLocsolas(testCase,capacity) == result, "teszteset: " + testCase + ", capacity: " + capacity )
    }
  }

  "differences" should "ki kellene adnia a különbségek listáját" in {
    def generateTestCase: (List[Int], List[Int] ) = Random.nextInt(10) match {
      case 0 => (Nil,Nil)
      case _ => {
        val (tail, differences) = generateTestCase
        val head = Random.nextInt(10)
        tail match {
          case Nil => (head::Nil, Nil)
          case head2 :: tail2 => (head::tail, (head2-head)::differences)
        }
      }
    }
    for( i <- 1 to 1000 ) {
      val (testCase, result) = generateTestCase
      assert(Gyakorlat.differences(testCase) == result , "teszteset: " + testCase )
    }
  }

  "hegymaszas" should "vissza kéne adja a pozitív különbségek összegét" in {
    def generateTestCase: (List[Int], Int) = Random.nextInt(5) match {
      case 0 => (Nil, 0)
      case _ => {
        val head = Random.nextInt(10) - 5
        val (tail, sum) = generateTestCase
        tail match {
          case Nil => ( head::Nil, 0 )
          case head2 :: tail2 => ( head::tail, if(head < head2) (head2-head) + sum else sum )
        }
      }
    }
    for( i <- 1 to 1000 ) {
      val (testCase, result) = generateTestCase
      testCase match {
        case head :: _ if head > 0 => assert( Gyakorlat.hegymaszas(testCase) == result + head, "input: "+testCase )
        case _ => assert( Gyakorlat.hegymaszas(testCase) == result, "input: "+testCase )
      }
    }
  }

  "csillagrendszer" should "true-t kéne visszaadjon, ha az input true,true,..,true,false,false,..,false" in {
    def generateConstant(v: Boolean) : List[Boolean] = Random.nextInt(5) match {
      case 0 => Nil
      case _ => v :: generateConstant( v )
    }
    def generateRandom: List[Boolean] = Random.nextInt(5) match {
      case 0 => Nil
      case _ => Random.nextBoolean() :: generateRandom
    }
    val kozetPrefix = generateConstant(true)
    val gazPrefix = generateConstant(false)
    val crazySuffix = if (gazPrefix.isEmpty) Nil else generateRandom
    assert( Gyakorlat.valodiCsillagrendszer( kozetPrefix ::: gazPrefix ::: crazySuffix ) == (!crazySuffix.contains(true)), "teszteset: " + (kozetPrefix ::: gazPrefix ::: crazySuffix ))
  }
}

